{
  "cells": [
    {
      "cell_type": "markdown",
      "id": "d6161252-e815-48d5-a2a3-5444c12e3b49",
      "metadata": {},
      "source": [
        "# 297. Serialize and Deserialize Binary Tree\n",
        "\n",
        "**Difficulty:** Hard\n",
        "\n",
        "**Link to Problem:** [To see the Serialize and Deserialize Binary Tree problem on LeetCode, click here!](https://leetcode.com/problems/serialize-and-deserialize-binary-tree/)\n",
        "\n",
        "---\n",
        "Serialization is the process of converting a data structure or object into a sequence of bits so that it can be stored in a file or memory buffer, or transmitted across a network connection link to be reconstructed later in the same or another computer environment.\n",
        "\n",
        "Design an algorithm to serialize and deserialize a binary tree. There is no restriction on how your serialization/deserialization algorithm should work. You just need to ensure that a binary tree can be serialized to a string and this string can be deserialized to the original tree structure.\n",
        "\n",
        "**Clarification:** The input/output format is the same as how LeetCode serializes a binary tree. You do not necessarily need to follow this format, so please be creative and come up with different approaches yourself.\n",
        "\n",
        "**Constraints:**\n",
        "- The number of nodes in the tree is in the range $[0, 10^4]$.\n",
        "- `-1000 <= Node.val <= 1000`"
      ]
    },
    {
      "cell_type": "markdown",
      "id": "fe9e71ae-2099-4eb9-b7a1-ec795ce92c76",
      "metadata": {},
      "source": [
        "## Probelm Explanation:\n",
        "\n",
        "TThe problem is to design an algorithm to serialize and deserialize a binary tree. Serialization is the process of converting a data structure or object into a sequence of characters or bytes so that it can be easily stored in a file, transmitted over a network, or otherwise persisted. Deserialization is the reverse process of converting the serialized data back into the original data structure.\n",
        "\n",
        "In the context of this problem:\n",
        "\n",
        "1. **Serialization**: You are given a binary tree, and your task is to convert it into a string representation such that you can later recreate the same binary tree from this string. The format of serialization is flexible, but it should allow you to reconstruct the original binary tree accurately.\n",
        "\n",
        "2. **Deserialization**: Given a serialized string, you need to reconstruct the binary tree it represents, ensuring that it is identical to the original tree.\n",
        "\n",
        "Here are some key points to consider in solving this problem:\n",
        "\n",
        "- The input can include any valid binary tree, including trees with nodes having integer values within the range [-1000, 1000].\n",
        "- You don't necessarily need to follow a specific format for serialization, but it should be designed in a way that allows unambiguous deserialization.\n",
        "- The goal is to serialize the tree structure and its values and then deserialize it back into the same structure and values.\n",
        "\n",
        "A common approach for serialization is to use a traversal method like preorder traversal, where you visit nodes in the order: root, left subtree, right subtree. This way, you can serialize the tree into a string, and during deserialization, you can reconstruct the tree by parsing the string in the same order.\n",
        "\n",
        "## Solution:\n",
        "Here's a Python function to implement this algorithm:"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 12,
      "id": "aabf7095-07dc-4ab7-876c-ec7213ed0de5",
      "metadata": {
        "trusted": true
      },
      "outputs": [],
      "source": [
        "# Definition for a binary tree node.\n",
        "class TreeNode(object):\n",
        "    def __init__(self, x):\n",
        "        self.val = x\n",
        "        self.left = None\n",
        "        self.right = None\n",
        "\n",
        "class Codec:\n",
        "    def serialize(self, root):\n",
        "        serialized  = []\n",
        "\n",
        "        def dfs(node):\n",
        "            if not node:\n",
        "                # If the node is None, represent it as \"Null\" in the serialized string\n",
        "                serialized .append(\"Null\")\n",
        "                return\n",
        "            # Convert the node's value to a string and add it to the serialized string\n",
        "            serialized .append(str(node.val))\n",
        "            dfs(node.left)\n",
        "            dfs(node.right)\n",
        "\n",
        "        dfs(root)\n",
        "        # Join the serialized values with \",\"\n",
        "        return \",\".join(serialized )\n",
        "\n",
        "    def deserialize(self, data):\n",
        "        vals = data.split(\",\")\n",
        "        self.i = 0\n",
        "\n",
        "        def dfs():\n",
        "            if vals[self.i] == \"Null\":\n",
        "                self.i += 1\n",
        "                return None\n",
        "            # Convert the value to an integer to create a new node\n",
        "            node = TreeNode(int(vals[self.i]))\n",
        "            self.i += 1\n",
        "            node.left = dfs()\n",
        "            node.right = dfs()\n",
        "            return node\n",
        "\n",
        "        return dfs()"
      ]
    },
    {
      "cell_type": "markdown",
      "id": "ece39324-cd9e-47d7-8677-a001878494d9",
      "metadata": {},
      "source": [
        "## Explanation:\n",
        "\n",
        "Here's a step-by-step explanation of the code:\n",
        "\n",
        "1. **Definition of TreeNode**:\n",
        "   - The code defines a simple `TreeNode` class that represents a node in a binary tree. Each `TreeNode` has three attributes:\n",
        "     - `val`: The integer value stored in the node.\n",
        "     - `left`: A reference to the left child node (or None if there is no left child).\n",
        "     - `right`: A reference to the right child node (or None if there is no right child).\n",
        "   - This class allows you to create binary tree nodes with integer values and left and right child references.\n",
        "\n",
        "2. **Serialization (serialize method)**:\n",
        "   - The `serialize` method takes a `root` node as input, which is the root of the binary tree that needs to be serialized.\n",
        "   - It initializes an empty list `serialized` to store the serialized elements.\n",
        "   - The `dfs` (depth-first search) function is defined within the `serialize` method to perform a preorder traversal of the binary tree.\n",
        "   - In the `dfs` function:\n",
        "     - If the current `node` is `None` (i.e., a leaf node or a child of a leaf node), it appends the string \"Null\" to the `serialized` list to represent the absence of a node.\n",
        "     - If the current `node` is not `None`, it appends the string representation of the `node.val` to the `serialized` list and then recursively calls `dfs` on the left and right children.\n",
        "   - After the `dfs` traversal, the `serialized` list contains the serialized binary tree elements.\n",
        "   - The method returns a string obtained by joining the elements in the `serialized` list with commas.\n",
        "\n",
        "3. **Deserialization (deserialize method)**:\n",
        "   - The `deserialize` method takes a serialized string `data` as input, which represents a binary tree in the custom format.\n",
        "   - It splits the `data` string by commas to obtain a list of elements called `vals`.\n",
        "   - The method initializes an index `self.i` to 0. This index keeps track of the current position in the `vals` list during deserialization.\n",
        "   - The `dfs` function is defined within the `deserialize` method to perform the deserialization process using a recursive approach.\n",
        "   - In the `dfs` function:\n",
        "     - If the current element in `vals` is \"Null,\" it means there is no node at this position in the binary tree, so it returns `None`.\n",
        "     - If the current element is not \"Null,\" it converts the element to an integer to create a new `TreeNode` with that value.\n",
        "     - It then recursively calls `dfs` to set the left and right children of the current node.\n",
        "   - The method returns the root node of the reconstructed binary tree.\n",
        "\n",
        "Overall, this code demonstrates a way to serialize a binary tree into a string format and then deserialize it back into the original tree structure, allowing for the representation of both the tree structure and the values stored in the nodes."
      ]
    },
    {
      "cell_type": "markdown",
      "id": "a54ea5b7-c53f-46c0-a44f-ef769ac68244",
      "metadata": {},
      "source": [
        "## Test cases:\n",
        "\n",
        "Here's how you can use this solution:"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 13,
      "id": "355f5bfd-94e1-4c5b-9914-cc20775f4160",
      "metadata": {
        "trusted": true
      },
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "1,2,Null,Null,3,4,Null,Null,5,Null,Null\n"
          ]
        }
      ],
      "source": [
        "# Example 1: \n",
        "\n",
        "# Example usage:\n",
        "# Serialize a binary tree\n",
        "root = TreeNode(1)\n",
        "root.left = TreeNode(2)\n",
        "root.right = TreeNode(3)\n",
        "root.right.left = TreeNode(4)\n",
        "root.right.right = TreeNode(5)\n",
        "\n",
        "codec = Codec()\n",
        "serialized_tree = codec.serialize(root)\n",
        "print(serialized_tree)\n",
        "\n",
        "# Deserialize the serialized tree\n",
        "new_root = codec.deserialize(serialized_tree)"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 14,
      "id": "192912fb-9d12-4911-9a4f-bfd37043e11d",
      "metadata": {
        "trusted": true
      },
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "Null\n"
          ]
        }
      ],
      "source": [
        "# Example 2:\n",
        "\n",
        "# Example usage:\n",
        "# Serialize a binary tree\n",
        "root = None\n",
        "\n",
        "codec = Codec()\n",
        "serialized_tree = codec.serialize(root)\n",
        "print(serialized_tree)\n",
        "\n",
        "# Deserialize the serialized tree\n",
        "new_root = codec.deserialize(serialized_tree)"
      ]
    },
    {
      "cell_type": "markdown",
      "id": "7d9dd448-c5f5-4192-8aa5-7c70ec7f42e2",
      "metadata": {},
      "source": [
        "## Time and Space Complexity Analysis\n",
        "\n",
        "Let's analyze the time and space complexity of the provided code for serializing and deserializing a binary tree:\n",
        "\n",
        "1. **Serialization (serialize method)**:\n",
        "\n",
        "   - Time Complexity:\n",
        "     - The `serialize` method uses a depth-first traversal (preorder) of the binary tree.\n",
        "     - It visits each node exactly once.\n",
        "     - At each node, it performs constant-time operations (string conversion and list append).\n",
        "     - Therefore, the time complexity for serialization is O(N), where N is the number of nodes in the binary tree.\n",
        "\n",
        "   - Space Complexity:\n",
        "     - The space complexity for serialization includes the space used by the `serialized` list and the recursive call stack.\n",
        "     - The `serialized` list stores the serialized elements, and its space is proportional to the number of nodes.\n",
        "     - The recursive call stack depth is determined by the height of the binary tree, and in the worst case (completely unbalanced tree), it can be O(N).\n",
        "     - Therefore, the space complexity for serialization is O(N) due to the list and O(H) due to the recursive call stack, where H is the height of the tree. In most cases, the dominant factor is O(N).\n",
        "\n",
        "2. **Deserialization (deserialize method)**:\n",
        "\n",
        "   - Time Complexity:\n",
        "     - The `deserialize` method splits the serialized string into a list of elements, which takes O(N) time.\n",
        "     - The `dfs` function is a recursive function that visits each element in the list exactly once.\n",
        "     - At each element, it performs constant-time operations (string comparison, integer conversion, and recursive function calls).\n",
        "     - Therefore, the time complexity for deserialization is O(N).\n",
        "\n",
        "   - Space Complexity:\n",
        "     - The space complexity for deserialization includes the space used by the `vals` list and the recursive call stack.\n",
        "     - The `vals` list stores the elements from the serialized string, and its space is proportional to the number of nodes.\n",
        "     - The recursive call stack depth is determined by the height of the binary tree, and in the worst case (completely unbalanced tree), it can be O(N).\n",
        "     - Therefore, the space complexity for deserialization is O(N) due to the list and O(H) due to the recursive call stack, where H is the height of the tree. In most cases, the dominant factor is O(N).\n",
        "\n",
        "Overall, the provided code has a time complexity of O(N) for both serialization and deserialization and a space complexity of O(N) in the majority of cases (unless the tree is severely unbalanced, in which case the height H dominates the space complexity). It efficiently serializes and deserializes a binary tree while using a linear amount of memory."
      ]
    },
    {
      "cell_type": "markdown",
      "id": "037b210a-555b-4da3-8130-b74d1e4f33f6",
      "metadata": {},
      "source": [
        "## Challenging Exercises:\n",
        "\n",
        "1. **Efficient Serialization**:\n",
        "   Optimize the serialization process to minimize the length of the serialized string. Consider using techniques like binary encoding to represent node values more efficiently.\n",
        "\n",
        "2. **Custom Serialization Format**:\n",
        "   Modify the serialization and deserialization methods to use a custom format different from the one provided. Ensure that the new format allows for accurate reconstruction of the original binary tree."
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python (Pyodide)",
      "language": "python",
      "name": "python"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "python",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.8"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 5
}
